#include "polygon.h"
#include "insertion-sort.h"

#include <cmath>
#include <iostream>

using namespace std;

double distance(double xA, double yA, double xB, double yB)
{
   double delta_x_AB = xB - xA;
   double delta_y_AB = yB - yA;
   return sqrt(delta_x_AB*delta_x_AB + delta_y_AB*delta_y_AB);
}

double circumference(int n, double* x, double* y)
{
   if((n < 1) || (x == NULL) || (y == NULL))
   {
      cout << "Invalid data passed to function circumference in polygon.cpp.\n";
      return 0.0;
   }

   double sum = 0.0;
   for(int i = 0; i < n; i++)
   {
      int j = (i + 1) % n;  // if i is n-1, then j will be 0
      sum += distance(x[i], y[i], x[j], y[j]);
   }
   return sum;
 }

// m and c are passed by reference;
// after returning, *m and *c contain the coefficients of the line
// equation, y = mx + c, which connects point A(xA/yA) with B(xB/yB)
//
void get_line_equation(double xA, double yA, double xB, double yB, double* m, double* c)
{
   if((m == NULL) || (c == NULL))
   {
      cout << "Invalid address found in function get_line_equation (polygon.cpp).\n";
      return;
   }
   if(xA == xB)  // avoid division by zero
   {
      if(yA == yB)  // no line, but actually a single point
      {
         *m = 0.0;
         *c = yA;
         return;
      }
      xA = xA - 1.0e-09*(yB - yA);  // shift xA and xB unnoticeably
      xB = xB + 1.0e-09*(yB - yA);
   }

   *m = (yB - yA) / (xB - xA);  // determine the slope; write it to *m
   *c = yA - *m * xA;  // determine the offset; write it to *c
   return;
}

// precondition:
//    x[n] and y[n] are arrays containing the coordinates of n points
//    m[n] and c[n] are arrays with n (double) elements
//
// postcondition:
//    after the function terminates, the arrays m and c contain the
//    line equations for the edges of the polygon as follows:
//
//    for all i between 0 and n - 2, m[i] and c[i] describe the line
//    given by the points (x[i] / y[i]) and (x[i+1] / y[i+1])
//
//    m[n-1] and c[n-1] describe the line given by the points
//    (x[n-1] / y[n-1]) and (x[0] / y[0])
//
// note:
//    for an array a, the element with index i is accessed by a[i];
//    accordingly, the address of this element is given by &(a[i])
//
void get_all_line_equations(int n, double* x, double* y, double* m, double* c)
{
   if((n < 1) || (x == NULL) || (y == NULL) || (m == NULL) || (c == NULL))
   {
      cout << "Invalid data found in function get_all_line_equations (polygon.cpp).\n";
      return;
   }

   for(int i = 0; i < n; i++)
   {
      int j = (i + 1) % n;  // if i is n-1, then j will be 0
      get_line_equation(x[i], y[i], x[j], y[j], (&m[i]), (&c[i]));
   }
   return;
}

bool is_valid_polygon(int n, double* x, double* y)
{
   if((n < 1) || (x == NULL) || (y == NULL))
   {
      cout << "Invalid data passed to function is_valid_polygon in polygon.cpp.\n";
      return false;
   }
   if(n < 3)
   {
      cout << "A valid polygon needs at least three corners.\n";
      return false;
   }

   double edge_m[n];
   double edge_c[n];
   get_all_line_equations(n, x, y, edge_m, edge_c);

   // first presume that the polygon is valid
   // then check whether there are any intersections, making it invalid
   //
   bool is_valid = true;

   for(int i = 0; i < n - 2; i++)
   {
      int jmax = n - 1;
      if(i == 0) jmax = n - 2;  // intersection between lines 0 and n-1 is a corner

      for(int j = i + 2; j <= jmax; j++)  // i and i+1 intersect at a corner, which is OK
      {
         if(edge_m[i] != edge_m[j])  // case that the two lines do not have the same slope
         {
            // determine: where do line i and line j intersect?
            //
            double xint = (edge_c[j] - edge_c[i]) / (edge_m[i] - edge_m[j]);
            double yint = edge_m[i] * xint + edge_c[i];

            // is it outside (which is OK) or inside (making the
            // polygon invalid)? check by comparing xint to x
            // coordinates of the end points of line i, which are those
            // of point i and point i+1
            //
            double xmin_line_i, xmax_line_i;
            if(x[i] > x[i+1])
            {
               xmin_line_i = x[i+1];
               xmax_line_i = x[i];
            }
            else
            {
               xmin_line_i = x[i];
               xmax_line_i = x[i+1];
            }
            if((xmax_line_i >= xint) && (xint >= xmin_line_i))
            {
               is_valid = false;  // intersection is inside: invalid

               cout << "Attention: Invalid intersection was detected "
                    << "at (" << xint << " / " << yint << ") between edge "
                    << i << ",\n\ty = " << edge_m[i] << " x + " << edge_c[i]
                    << "\nand edge " << j << ",\n\ty = " << edge_m[j]
                    << " x + " << edge_c[j] << "\n";
            }
         }
         else  // consider the case that the slopes of the lines are the same
         {
            if(edge_c[i] == edge_c[j])
            {
               is_valid = false;  // twice the same edge: invalid

               cout << "Attention: The edges " << i << " and " << j
                    << " are the same; both are given by "
                    << "\n\ty = " << edge_m[i] << " x + " << edge_c[i]
                    << ".\n";
            }
         }
      }
   }
   return is_valid;
}

double area(int n, double* x, double* y)
{
   if((n < 1) || (x == NULL) || (y == NULL))
   {
      cout << "Invalid data passed to function area in polygon.cpp.\n";
      return 0.0;
   }

   // create a working copy of x and y; this is to make sure that the
   // originals are not affected by the sorting operations that will
   // need to be carried out
   //
   double xwork[n];
   double ywork[n];
   for(int i=0; i < n; i++)
   {
      xwork[i] = x[i];
      ywork[i] = y[i];
   }

   // now sort the points of the working copy according to the value of
   // their x coordinates
   //
   // after this function call, the points given by
   // (xwork[i] / ywork[i]) will be the same as (x[j] / y[j]), for
   // some value of j, and they will be arranged such that the elements
   // of the array work are sorted in ascending order
   //
   // depending on how you implement the algorithm discussed in the
   // lecture, sorting the x coordinates may be enough; in this case,
   // you can, but need not, replace this the line below with the
   // function call "insertion_sort(n, xwork)"
   //
   insertion_sort_cosort(n, xwork, ywork);

   // note that while (x[i] / y[i]) and (x[i+1] / y[i+1]) define an
   // edge of the polygon, after rearranging the points in xwork and
   // ywork, this will not apply to them
   //
   // hence, x and y, rather than xwork and ywork, need to be
   // considered to obtain the line equations for the polygon edges
   //
   double edge_m[n];
   double edge_c[n];
   get_all_line_equations(n, x, y, edge_m, edge_c);

   //
   // CODE MISSING HERE
   //
   // AREA NEEDS TO BE DETERMINED (E.G., AS DISCUSSED IN THE LECTURE)
   //

   return 0;
}
